#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)

"""
Script: Generate a summary of the QAR production status
---
- Read QAR directory
- Generate bokeh table plot
"""

import os
import shutil
import pandas as pd
import configparser
from datetime import date
from math import pi

from bokeh.io import save, output_file
from bokeh.models import LinearColorMapper
from bokeh.plotting import figure

"""
Generate a Bokeh table plot with tooltips
Each row is a QAR
"""
def table_plot(data, output_html, total):

	colors = ["#550b1d",  "#dddddd",  "#75968f"]
	mapper = LinearColorMapper(palette=colors, low=-1, high=1)

	yy = data.fname.unique()
	xx = data.test.unique()
	p = figure(title="C3S QAR status ({}) - {} of {} completed".format(
		       date.today().strftime('%d %b %Y'), total, len(yy)),
	     	   x_range=xx, y_range=yy,
	           x_axis_location="above", plot_width=1000, plot_height=7000,
	           tools='', toolbar_location='below',
	           tooltips=[('msg', '@msg')])

	p.toolbar.logo = None
	p.toolbar_location = None
	p.grid.grid_line_color = None
	p.axis.axis_line_color = None
	p.axis.major_tick_line_color = None
	p.axis.major_label_text_font_size = "7pt"
	p.axis.major_label_standoff = 0
	p.xaxis.major_label_orientation = pi / 3

	p.rect(x="test", y="fname", width=1, height=1,
	       source=data,
	       fill_color={'field': 'res', 'transform': mapper},
	       line_color=None)

	# Save plot
	output_file(output_html)
	save(p)


"""
OK/NOK to integers
"""
def ok_nok(val):
	if val == 'ok':
		return 1
	else:
		return -1

"""
Get information from QAR/ini result
"""
def parse_ini(ff):

	# There is empty
	res = {}
	metadata = {}

	if os.path.exists(ff):
		ini = configparser.ConfigParser()
		ini.read(ff)
		res = {}
		metadata = {}

		for s in ini.sections():
			if 'result' in s:
				res[s.replace('_result', '')] = ok_nok(ini[s]['res'])	
				metadata[s.replace('_result', '')] = '{}: {}'.format(ini[s]['res'], ini[s]['msg'])

	return res, metadata

"""
Generate QAR report main
"""		    
def main():

	# Get all ini files
	script_dir = os.path.dirname(os.path.realpath(__file__))
	conf_dir = os.path.join(script_dir, 'conf')
	
	# Results dir
	res_dir = os.path.join(script_dir, 'results')
	
	# Cleanup directory
	shutil.rmtree(res_dir)
	os.mkdir(res_dir)
	os.chdir(res_dir)

	# Get all ini files	
	os.system("wget -q ftp://bscesftp.bsc.es/jsala/cds_data_checks/*.ini")

	# For every result available
	tests = [
		'file_format', 'standard_compliance', 
		'temporal_completeness', 'temporal_consistency',
		'spatial_completeness', 'spatial_consistency', 'valid_ranges'
	]

	df = pd.DataFrame(index=[], columns=['fname', 'test', 'res'])
	tt = 0
	for f in os.listdir(conf_dir):
		ini_res = os.path.join(res_dir, f)
		rs, mdata = parse_ini(ini_res)

		# Data available
		if rs != {}:
			tt+=1
			for k,v in rs.items():			
				df=df.append({'fname':f.replace('.ini', ''), 'test': k, 'res': v, 'msg': mdata[k]}, 
					ignore_index=True)
		else:
			for c in tests:			
				df=df.append({'fname':f.replace('.ini', ''), 'test': c, 'res': 0, 'msg': 'data unavailable'}, 
					ignore_index=True)

	# Plot html table
	print(df.to_csv())
	table_plot(df, os.path.join(res_dir, 'qar_status.html'), tt)

if __name__ == "__main__":
    main()