#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)

import sys
import logging
import datetime
import subprocess
import os
import configparser
import traceback

# Local imports
from utils.utils_files import *
from utils.utils_pdf import *

"""
Class: Versioning and messages of the wrapper
"""
class checker_body:

    def __init__(self, name):
        self.name = name
        self.version = "0.0.3"
        self.help_mssg()

    def __version__(self):
        
        string = """        
          _____     ____     _____
         |  __ \   / __ \   / ____|
         | |  | | | |  | | | |
         | |  | | | |  | | | |
         | |__| | | |__| | | |____
         |_____/   \___\_\  \_____|
        
          Data Quality Checker - version {} (C3512 project)
        """.format(self.version)
        
        print(string.replace('\t', ''))

    def help_mssg(self):
        string = "\n#"+ "#" * 80 + "\n#\n"
        string += "#"+ " " * 2+"Data Quality Checker Help message - How to run the DQC:\n#\n"
        string += "#"+ " " * 4+"$ python checker.py <config_file.ini>\n#\n"
        string += "#" + "#" * 80 + "\n"

        self.help = string

"""
Class: Main interface to the checks
---
- Reads config file as unique commandline operator
- Loops over the different checks/sections
"""
def main(args):
    dqc = checker_body("testing")
    dqc.__version__()

    # Help display
    try:
        if len(args) > 1:
            raise Exception('Too many arguments')
        if (args[0] in ["--help","-h","help","sos","commands"]):
            raise Exception('Help message')
    except:
        print(dqc.help)
        sys.exit()        

    # Read configuration and check mandatory fields in the general section   
    config = None
    config_fname = os.path.basename(args[0])
    try:
        if os.path.exists(args[0]):  
            config = configparser.ConfigParser()
            config.read(args[0])
        else:
            raise Exception('Configuration file does not exist')

        # General settings
        path_files = check_param_config(config, 'general', 'input')
        fpattern = check_param_config(config, 'general', 'fpattern')
        log_dir = check_param_config(config, 'general', 'log_dir')
        res_dir = check_param_config(config, 'general', 'res_dir')
        forms_dir = check_param_config(config, 'general', 'forms_dir')
        
        # Dataset settings
        cds_dataset = check_opt_param_config(config, 'dataset', 'cds_dataset')
        cds_variable = check_opt_param_config(config, 'dataset', 'cds_variable')
        varname = check_opt_param_config(config, 'dataset', 'variable')
        dt = check_datatype_config(config, 'dataset', 'datatype')
        nbp = check_opt_param_config(config, 'dataset', 'numberofpoints')
        cz = int(check_opt_param_config(config, 'dataset', 'chunksize', default=-1)) # default is no chunks

        # Directory checks
        if not os.path.exists(path_files):  
            raise Exception('Input directory does not exist. Check configuration file section [general]')
        if not log_dir:  
            raise Exception('Log directory not defined. Check configuration file section [general], value log_dir')
        if not res_dir:  
            raise Exception('Results directory not defined. Check configuration file section [general], value res_dir')

        # Init logging and list files
        log_file = initiate_logging(log_dir, path_files, config_fname)
        files = dir_or_file(path_files, fpattern)

        # Not needed anymore
        config.remove_section('general')
        config.remove_section('dataset')

    except Exception as e:
        print('ERROR: {}'.format(e))
        sys.exit()
        
    # Skip if zip file exists
    zipfname = get_zip_fname(res_dir, args[0])
    if os.path.exists(zipfname):
        print('INFO: Skipping {}'.format(zipfname))
        sys.exit()    

    # For each section/test
    files2zip = []
    err = False

    for s in config.sections(): 
        
        string = "." + "." * 80 + "\n"
        string += "Data Quality Checker [check={}]\n".format(s)        
        string += "." + "." * 80 + "\n"
        print(string)        

        # Each section is a different test available  
        try:
            test_res = None

            ## TEST 00
            if s == 'file_format':
                import test00_file_format as test
                file_format_ck = test.file_format_check(files)
                test_res = file_format_ck.main()

            ## TEST 01
            elif s == 'standard_compliance':            
                import test01_standard_compliance as test            
                standard_ck = test.standard_compliance_test(files, forms_dir, cds_dataset, cds_variable)
                test_res = standard_ck.main()

            ## TEST 02
            elif s == 'spatial_completeness':                                            
                mf = check_opt_param_config(config, s, 'mask_file', default='') # default is without mask
                mv = check_opt_param_config(config, s, 'mask_var', default='lsm') # default is lsm (land_sea_mask)
                import test02_spatial_completeness as test
                spatial_completeness_ck = test.spatial_completeness_test(files, varname, dt, nbp, cz, mf, mv)
                test_res = spatial_completeness_ck.main()                               

            ## TEST 03
            elif s == 'temporal_completeness': 
                origin = check_param_config(config, s, 'origin')
                system = check_param_config(config, s, 'system')
                is_s = check_opt_param_config(config, s, 'is_seasonal', default='no') # default is not seasonal      
                import test03_temporal_completeness as test            
                temporal_completeness_ck = test.temporal_completeness_test(files, forms_dir, 
                    cds_dataset, cds_variable, origin, system, is_s)
                test_res = temporal_completeness_ck.main()

            ## TEST 04
            elif s == 'spatial_consistency':              
                grid_int = check_param_config(config, s, 'grid_interval')
                grid_typ = check_param_config(config, s, 'grid_type')
                grid_map = check_opt_param_config(config, s, 'grid_mapping_name', default='') # default is empty      
                import test04_spatial_consistency as test
                spatial_consistency_ck = test.spatial_consistency_test(files, grid_typ, grid_int, grid_map)
                test_res = spatial_consistency_ck.main()
            
            ## TEST 05
            elif s == 'temporal_consistency':
                time_s = check_param_config(config, s, 'time_step')
                time_g = check_param_config(config, s, 'time_granularity')
                import test05_temporal_consistency as test
                temporal_consistency_ck = test.temporal_consistency_test(files, time_s, time_g)  
                test_res = temporal_consistency_ck.main()          

            ## TEST 06
            elif s == 'valid_ranges':
                vmin = check_param_config(config, s, 'valid_min')
                vmax = check_param_config(config, s, 'valid_max')
                import test06_valid_ranges as test                
                valid_ranges_ck = test.valid_ranges_test(res_dir, config_fname, files, varname, vmin, vmax, 
                    dt, nbp, cz, forms_dir, cds_dataset, cds_variable)
                test_res = valid_ranges_ck.main()       
                files2zip.append(test_res['plot'])
                
            else:
                raise Exception('Test {} is not available in the current checker implementation. Please check your config syntax')
        
        # Exceptions/Errors
        except KeyboardInterrupt:
            print_progress(100, 100)
            err=True
            print('Cancelled, exiting...')
            exit()
        except:            
            print_progress(100, 100)
            err=True
            print('ERROR: Processing section {}'.format(s))
            print('ERROR: {}'.format(traceback.format_exc()))

        # Append result
        finally:  
            print()          
            if test_res != None:
                config.add_section(s+'_result')
                config.set(s+'_result', 'res', test_res['res'])
                config.set(s+'_result', 'msg', test_res['msg'])
                                        
                if test_res['res'] == 'ok':
                    print('INFO: {} -> {}'.format('PASS', test_res['msg']))
                else:                    
                    print('ERROR: {} -> {}'.format('FAIL', test_res['msg']))

    try:
        # Save TXT result object
        res_file = save_result(config, res_dir, config_fname)

        # Save PDF report (result + log)        
        if os.path.exists(res_file) and os.path.exists(log_file) and not(err):
            for s in config.sections():
                if not('result' in s):
                    files2zip.append(save_pdf_report(s, res_file, log_file))


    except:
        print('ERROR: Could not save results. Check res_dir and log_dir permissions')

    # Zip all results
    if not(err):         
        zip_results(zipfname, files2zip)


if __name__ == "__main__":
    args = sys.argv[1:]
    sys.exit(main(args))